/* this is queries.c file.
 * this is part of the libspopc library sources
 * copyright  2002 Benoit Rouits <brouits@free.fr>
 * released under the terms of GNU LGPL
 * (GNU Lesser General Public Licence).
 * libspopc offers simple API for a pop3 client (MTA).
 * See RFC 1725 for pop3 specifications.
 * more information on http://brouits.free.fr/libspopc/
 */

#ifdef WIN32
#include <winsock.h>
#else
#include <sys/socket.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <sys/time.h>
#include <sys/types.h>
#ifndef WIN32
#include <unistd.h>
#endif

#include "libspopc.h"

#ifdef USE_SSL

int pop3_recv (pop3sock_t sock, char* buf, int len){
	return sock->ssl?SSL_read(sock->ssl,buf,len):recv(sock->sock,buf,len,0);
}


int pop3_send (pop3sock_t sock, char* buf, int len){
	return sock->ssl?SSL_write(sock->ssl,buf,len):send(sock->sock,buf,len,0);
}

#else

int pop3_recv (pop3sock_t sock, char* buf, int len){
	return recv(sock,buf,len,0);
}

int pop3_send (pop3sock_t sock, char* buf, int len){
	return send(sock,buf,len,0);
}

#endif

/* enable timeout sockets with 'select' */
int timedrselect(fd_set *rfds, int max, int time){
struct timeval timeout;
timeout.tv_sec = time;
timeout.tv_usec = 0;

   if(select(max, rfds, NULL, NULL, &timeout) <= 0){
      perror("timeout reached\n");
      return(1);
   }
   return(0);
}

/* prepare the socket to be read */
int timedrread(pop3sock_t sock, int time) {
#if 0
fd_set rfds;
#ifdef USE_SSL
#define SOCK (sock->sock)
   if (sock->ssl) return 1;
#else
#define SOCK (sock)
#endif
   FD_ZERO(&rfds);
   FD_SET(SOCK, &rfds);
   timedrselect(&rfds, SOCK + 1, time);
   return FD_ISSET(SOCK, &rfds);
#else
   sock = 0; time=0;
   return 1;
#endif
}

char* pop3_query(pop3sock_t sock, const char* query){
/* performs a simple pop query and returns server's <512 bytes response */
    int r=0;
    int bytes=0;
    char* buf=NULL;

    r=pop3_send(sock,(char *)query,strlen(query));
    if(r==-1){
		perror("pop3_query.send");
        return(NULL);
    }
    buf=(char*)malloc(POPBUF+2); /* 512 chars + '\0' */
    if(!buf){
        perror("pop3_query.malloc");
        return(NULL);
    }
    do {
        bytes+=pop3_recv(sock, buf+bytes, POPBUF-bytes);
        if(buf[0]=='\r' && buf[1]=='\n') {
            break;
        }
    } while (buf[bytes-2]!='\r' || buf[bytes-1]!='\n');
    buf[bytes+1]='\0';
    return(buf);
}

char* pop3_user(pop3sock_t sock, const char* name){
/* performs "USER" pop query and returns server's <512 bytes response */
char query[POPBUF]; /* total "USER ****your_name****\n" is < 512  */

	snprintf(query,POPBUF,"USER %s\r\n",name);
	return(pop3_query(sock,query));
}

char* pop3_pass(pop3sock_t sock, const char* pw){
/* performs "PASS" pop query and return server's <512 bytes response */
char query[POPBUF]; /* total "PASS ****your_pass****\n" is <512 */

	snprintf(query,POPBUF,"PASS %s\r\n",pw);
	return(pop3_query(sock,query));
}

char* pop3_quit(pop3sock_t sock){
/* performs "QUIT" pop query and returns server's <512 bytes response */
char query[]="QUIT\r\n";

	return(pop3_query(sock,query));
}

char* pop3_stat(pop3sock_t sock){
/* performs "STAT" pop query and returns server's <512 bytes response */
char query[]="STAT\r\n";

	return(pop3_query(sock,query));
}

char* recv_rest(pop3sock_t sock, char* buf, int cursize, int bufsize){
/* recv rest of data through sock, given a cs  pre-filled buffer sized of bs.
* end of data is assumed when data has a "\r\n.\r\n" string
* recv() is TCPBUFLEN  bytes stepped, SUGGEST any good value...
* Warning: after calling this function, buf must never be used again
* -not even for a free(buf)- since it may be reallocated. use the return
* value instead (usage example: buf=recv_rest(sock,buf,cs,bs); free(buf); */
char* ret = NULL;
char* cur = NULL; /* current position ready to receive */
int total;    /* total received */
	if(!buf){
		return(NULL);
	}
	total = cursize;
	cur = buf;
	if(cursize == bufsize){
		ret=(char*)realloc(buf,bufsize+1);
		if(!ret){
			perror("recv_rest.realloc");
            free(buf);
			return NULL;
		}
        cur = buf = ret;
	}
	cur[total]='\0';
	while(!dotline(buf)){
		if (total >= (bufsize - TCPBUFLEN)){
			ret = (char*)realloc(buf, (bufsize *=2) +1);
		}
		if(!ret){
			perror("recv_rest.realloc");
            free(buf);
			return NULL;
		}
		buf = ret;
		cur = buf + total;
		if(timedrread(sock, SOCKET_TIMEOUT)){
			cursize=pop3_recv(sock, cur, TCPBUFLEN);
			if (cursize < 0){ /* socket error */
				perror("Socket Error");
				free(buf);
				return(NULL);
			} else if (cursize == 0){ /* connection closed */
				perror("Connection closed by peer");
				free(buf);
				return(NULL);
			}
		}else{
			printf("timeout reached\n");
			return(buf);
		}
		if (cursize > 0){
			total+=cursize;
			cur[cursize] = '\0';
		}else{
			cursize = 0;
		}
	}
	cur[cursize] = '\0';
	return(buf);
}
      
char* pop3_list(pop3sock_t sock, int id){
/* performs a "LIST" pop query and returns server's (long) response */
int r;
char query[POPBUF]; /* total query "LIST ID\n" string is < 512 */
char* buf;

	if(id>0){
		snprintf(query,POPBUF,"LIST %d\r\n",id);
	}else{
		snprintf(query,POPBUF,"LIST\r\n");
	}
	r=pop3_send(sock,query,strlen(query));
	if(r==-1){
		perror("pop3_list.send");
		return(NULL);
	}
	/* now prepare a first short 512 bytes recv() */
	/* it might be now enough for recv() from "LIST X" */
	buf=(char*)malloc(POPBUF +1); /* 512 chars + '\0' */
	if(!buf){
		perror("pop3_list.malloc");
		return(NULL);
	}
	r=pop3_recv(sock,buf,POPBUF);
   if(r < 0){ /* socket error */
      perror("Socket Error");
      free(buf);
      return(NULL);
   }
   buf[r]='\0';
	if(id>0){/* +OK id size */
		return(buf); /* 512 bytes are enough as say RFC 1939 */
	}
	/* else : +OK X messages (YYY octets)\n id size\n... */
	if(pop3_error(buf)){
		return(buf);
	}
	return(recv_rest(sock,buf,r,POPBUF));
}


char* pop3_retr(pop3sock_t sock, int id){
int r;
char query[POPBUF];
char *buf;

   snprintf(query, POPBUF, "RETR %d\r\n", id);
   r=pop3_send(sock, query, strlen(query));
   if(r == -1){
      perror("pop3_retr.send");
      return(NULL);
   }
   buf=(char*)malloc(POPBUF +1);/* 512 chars + '\0' */
   if(!buf) {
      perror("pop3_retr.malloc");
      return(NULL);
   }
   if(timedrread(sock, SOCKET_TIMEOUT)){
      r=pop3_recv(sock, buf, POPBUF); 
      if(r < 0){ /* socket error */
         perror("Socket Error");
         free(buf);
         return(NULL);
      }
   }else{
      perror("timeout reached\n");
      free(buf);
      return(NULL);
   }
   if(pop3_error(buf)){
      buf[r+1] = '\0';
      return(buf); /* 512 are enough as say RFC 1939 */
   }
   if (r > 0) {
      return(recv_rest(sock, buf, r, POPBUF));
   }else{
      perror("nothing returned\n");
      free(buf);
      return(NULL);
   }
}

char* pop3_dele(pop3sock_t sock, int id){
/* performs a "DELE" pop query and returns server's <512 bytes response */
char query[POPBUF]; /* total "DELE X\n" string < 512 */

	if(id<=0){
		return(NULL);
	}
	snprintf(query,POPBUF,"DELE %d\r\n",id);
	return(pop3_query(sock,query));
}

char* pop3_noop(pop3sock_t sock){
/* performs a "NOOP" pop query and returns server's <512 bytes response */
char query[]="NOOP\r\n";
	
	return(pop3_query(sock,query));
}

char* pop3_rset(pop3sock_t sock){
/* performs a "RSET" pop query and returns server's <512 bytes response */
char query[]="RSET\r\n";

	return(pop3_query(sock,query));
}

char* pop3_top(pop3sock_t sock, int id, int lines){
/* performs a "TOP" pop query and returns server's (long) response */
int r;
char query[POPBUF]; /* total "TOP X Y\n" is < 512 */
char* buf;

	snprintf(query,POPBUF,"TOP %d %d\r\n",id,lines);
	r=pop3_send(sock,query,strlen(query));
	if(r==-1){
		perror("pop3_top.send");
		return(NULL);
	}
	/* prepare first recv() of 512 bytes */
	buf=(char*)malloc(POPBUF +1); /* 512 chars + '\0' */
	if(!buf){
		perror("pop3_top.malloc");
		return(NULL);
	}
	r=pop3_recv(sock,buf,POPBUF);
   if(r < 0){ /* socket error */
      perror("Socket Error");
      free(buf);
      return(NULL);
   }
   buf[r]='\0';
   
	if(pop3_error(buf)){
		return(buf); /* 512 bytes are enough as say RFC 1939 */
	}
	return(recv_rest(sock,buf,r,POPBUF));
}

char* pop3_uidl(pop3sock_t sock, int id){
/* performs a "UIDL" pop query and returns server's (long) response */
int r;
char query[POPBUF]; /* total "UIDL X\n" is < 512 */
char* buf;

	if(id>0){
		snprintf(query,POPBUF,"UIDL %d\r\n",id);
	}else{
		snprintf(query,POPBUF,"UIDL\r\n");
	}
	r=pop3_send(sock,query,strlen(query));
	if(r==-1){
		perror("pop3_uidl.send");
		return(NULL);
	}
	/* prepare first 512 bytes for recv() */
	/* i hope this is also enough for the 'one line' short response */
	buf=(char*)malloc(POPBUF +1); /* 512 chars + '\0' */
	if(!buf){
		perror("pop3_uidl.malloc");
		return(NULL);
	}
	memset(buf,0,POPBUF);
	r=pop3_recv(sock,buf,POPBUF);
   if (r < 0){ /* socket errorr */
      perror("Socket Error");
      free(buf);
      return(NULL);
   }
   buf[r]='\0';
   
	if(id>0){/* +OK id sig */
		/* anyway, we return the short buf, error or not */
		return(buf); /* 512 are enough as say RFC 1939 */
	}
	/* else : +OK\n id sig\nid sig\nid sig\n... */
	if(pop3_error(buf)){
		return(buf); /* hope error msg were < 512 bytes */
	}
	return(recv_rest(sock,buf,r,POPBUF));
}

char* pop3_apop(pop3sock_t sock, const char* name, const char* digest){
/* performs a "APOP" secure query and returns server's <512 bytes response */
char query[POPBUF]; /* total "APOP name digest\n" is < 512 */

	snprintf(query,POPBUF,"APOP %s %s\r\n",name,digest);
	return(pop3_query(sock,query));
}
